from datetime import datetime, timedelta
from typing import Dict, List, Type, TypedDict, Any
import requests
import aiohttp

from .errors import *

class EstimatedDiameterDetails(TypedDict):
    """
    Equals to 'estimated_diameter_max'
    and 'estimated_diameter_min' on
    API response
    """
    estimated_diameter_max: float
    estimated_diameter_min: float

class EstimatedDiameter(TypedDict):
    """
    Equals to 'estimated_diameter' on
    API response
    """
    kilometers: EstimatedDiameterDetails
    meters: EstimatedDiameterDetails
    miles: EstimatedDiameterDetails
    feet: EstimatedDiameterDetails

class RelativeVelocity(TypedDict):
    """
    Equals to 'relative_velocity' on
    API response
    """
    kilometers_per_second: str
    kilometers_per_hour: str
    miles_per_hour: str

class MissDistance(TypedDict):
    """
    Equals to 'miss_distance' on
    API response
    """
    astronomical: str
    lunar: str
    kilometers: str
    miles: str

class CloseApproachData(TypedDict):
    """
    Equals to 'close_approach_data' on
    API response
    """
    close_approach_date: str
    close_approach_date_full: str
    epoch_date_close_approach: int
    epoch_date_close_approach: int
    relative_velocity: RelativeVelocity
    miss_distance: MissDistance
    orbiting_body: str

class Asteroid(TypedDict):
    """
    Equals to an asteroid on API
    response
    """
    links: Dict[str, str]
    id: str
    neo_reference_id: str
    name: str
    nasa_jpl_url: str
    absolute_magnitude_h: float
    estimated_diameter: EstimatedDiameter
    is_potentially_hazardous_asteroid: bool
    close_approach_data: List[CloseApproachData]
    is_sentry_object: bool


class Links(TypedDict, total=False):
    """
    Equals to 'links'
    """
    next: str
    prev: str
    self: str

class NeoWsFeedResponse(TypedDict):
    """
    Equals to /feed response
    """
    links: Links
    element_count: int
    near_earth_objects: Dict[str, List[Asteroid]]

class OrbitClass(TypedDict):
    """
    Equals to 'orbit_class' in
    OrbitralData
    """
    orbit_class_type: str
    orbit_class_description: str
    orbit_class_range: str

class OrbitralData(TypedDict):
    """
    Equals to 'orbitral_data' in
    /browse asteroid response
    """
    oribt_id: str
    orbit_determination_date: str
    first_observation_date: str
    last_observation_date: str
    data_arc_in_days: int
    observations_used: int
    orbit_uncertainty: str
    minimum_orbit_intersection: str
    jupiter_tisserand_invariant: str
    epoch_osculation: str
    eccentricity: str
    semi_major_axis: str
    inclination: str
    ascending_node_longitude: str
    orbital_period: str
    perihelion_distance: str
    perihelion_argument: str
    aphelion_distance: str
    perihelion_time: str
    mean_anomaly: str
    mean_motion: str
    equinox: str
    oribt_class: OrbitClass

class NeoWsBrowseAsteroid(Asteroid, TypedDict):
    """
    Equals to an asteroid in /browse
    """
    orbitral_data: OrbitralData

class Page(TypedDict):
    """
    Refers to 'page' in /browse
    API response
    """
    size: int
    total_elements: int
    total_page: int
    number: int

class NeoWsBrowseResponse(TypedDict):
    links: Links
    page: Page
    near_earth_objects: List[NeoWsBrowseAsteroid]

class Validator:
    """
    Will be called to validate the options provided
    by the developer
    """
    @classmethod
    def validate(cls, options: Dict[str, Any], allowed_keys: List[str]) -> Dict[str, str]:
        # checking keys
        if "start_date" not in options.keys() and not "end_date" in options.keys():
            raise InvalidKey("You have to provide a 'start_date' or 'end_date'")

        # checking types
        for key in allowed_keys:
            if key in options.keys() and not isinstance(key, datetime):
                value = options[key]
                if not isinstance(value, datetime):
                    raise TypeError(f"'{key}' must be 'datetime.datetime', got '{value.__class__.__name__}'")
        
        # checking if dates are valid
        if options["end_date"] - timedelta(days=7) < datetime(1900, 1, 1):
            raise InvalidDate("'start_date' has a default value of 7 days before 'end_date', however, this value is before than Jan 1, 1. Please, set a 'start_date' or make 'end_date' be after Jan 8, 1900")
        elif options.get("start_date") and options["start_date"] < datetime(1900, 1, 1) or options["end_date"] < datetime(1900, 1, 1):
            raise InvalidDate("'start_date' and/or 'end_date' must be after Jan 1, 1900")

        # creating a start_date in case it doesnt exist
        if "end_date" in options.keys() and not "start_date" in options.keys():
            options["start_date"] = options["end_date"] - timedelta(days=7)

        # continue checking dates
        if options["start_date"] > options["end_date"]:
            raise InvalidDate("'start_date' can not be after 'end_date'")
        elif options["start_date"] + timedelta(days=7) < options["end_date"]:
            raise InvalidDate("'start_date' can is limited to 7 days before 'end_date' or minus")

        # stringify dates
        if options.get("end_date"):
            options["end_date"] = options["end_date"].strftime("%Y-%m-%d")
        options["start_date"] = options["start_date"].strftime("%Y-%m-%d")

        return options

class SyncNeoWs:
    """
    This class uses asynchronous programming
    syntax to make requests to the NeoWs
    web service and returns the API response.

    Descriptions of the methods are from
    the `NASA API Portal <https://api.nasa.gov/>`_.
    """
    def __init__(self, api_key: str) -> None:
        self._api_key = api_key
        self._allowed_keys = ["start_date", "end_date"]

    @property
    def api_key(self):
        """
        Returns the API key.
        """
        return self._api_key

    @property
    def allowed_keys(self):
        """
        Returns the allowed keys in list format.
        """
        return self._allowed_keys

    def get_neo_feed(self, options: Dict[str, Any]) -> NeoWsFeedResponse:
        """
        Retrieve a list of Asteroids based on
        their closest approach date to Earth.
        Here's a list of the allowed keys:
        
        **Allowed Keys**

            ========== ========================== ============================================
            Key        Type                       Function
            ========== ========================== ============================================
            start_date :class:`datetime.datetime` Sets a starting date for the asteroid search
            end_date   :class:`datetime.datetime` Sets a ending date for the asteroid search
            ========== ========================== ============================================

        **Example**

            .. code-block:: python3

                from nasawrapper import SyncNeoWs
                from datetime import datetime

                neows = SyncNeoWs("DEMO_KEY")
                result = neows.get_neo_feed({
                    "start_date": datetime(2010, 2, 3),
                    "end_date": datetime(2010, 2, 4)
                })
        """

        url = f"https://api.nasa.gov/neo/rest/v1/feed?api_key={self._api_key}"
        options = Validator.validate(options, self._allowed_keys)

        # building url
        for key, value in options.items():
            url += f"&{key}={value}"

        # making request
        request = requests.get(url)
        if request.status_code == 429:
            raise RateLimitError("You are being rate limited")
        elif request.status_code == 403:
            raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

        return request.json()

    def get_today_neo_feed(self) -> NeoWsFeedResponse:
        """
        Retrieve a list of Asteroids based on
        today's date in the current UTC.

        **Example**

            .. code-block:: python3

                from nasawrapper import SyncNeoWs

                neows = SyncNeoWs("DEMO_KEY")
                result = neows.get_today_neo_feed()
                print(result)
        """
        start_date = end_date = datetime.now()
        options = {
            "start_date": start_date,
            "end_date": end_date
        }

        return self.get_neo_feed(options)

    def get_neo_lookup(self, asteroid_id: int) -> Asteroid:
        """
        Lookup a specific Asteroid based on its
        NASA JPL small body (SPK-ID) ID. Find more
        about this at
        https://ssd.jpl.nasa.gov/tools/sbdb_query.html
        to get the overall data-set of asteroids and their ids.

        **Example**

            .. code-block:: python3

                from nasawrapper import SyncNeoWs

                neows = SyncNeoWs("DEMO_KEY")
                result = neows.get_neo_lookup(3542519)
                print(results)
        """
        if not isinstance(asteroid_id, int):
            raise TypeError(f"'asteroid_id' must be 'int', got {asteroid_id.__class__.__name__}")

        url = f"https://api.nasa.gov/neo/rest/v1/neo/{asteroid_id}?api_key={self._api_key}"
        

        # making request
        request = requests.get(url)
        if request.status_code == 404:
            raise NotFound(f"Asteroid of id '{asteroid_id}' could not be found")
        elif request.status_code == 429:
            raise RateLimitError("You are being rate limited")
        elif request.status_code == 403:
            raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

        return request.json()

    def get_neo_browse(self) -> NeoWsBrowseResponse:
        """
        Browse the overall Asteroid data-set.

        **Example**

            .. code-block:: python3

                from nasawrapper import SyncNeoWs

                neows = SyncNeoWs("DEMO_KEY")
                result = neows.get_neo_browse() # this may take some time, 
                                                # since the API's searching for all asteroids
                print(result)
        """
        url = f"https://api.nasa.gov/neo/rest/v1/neo/browse?api_key={self._api_key}"

        # making request
        request = requests.get(url)
        if request.status_code == 429:
            raise RateLimitError("You are being rate limited")
        elif request.status_code == 403:
            raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

        return request.json()

class AsyncNeoWs:
    """
    This class uses asynchronous syntax to 
    make requests to the NeoWs web service
    and returns the API response.

    Descriptions of the methods are from
    the `NASA API Portal <https://api.nasa.gov/>`_.
    """
    def __init__(self, api_key: str) -> None:
        self._api_key = api_key
        self._allowed_keys = ["start_date", "end_date"]

    @property
    def api_key(self):
        """
        Returns the API key.
        """
        return self._api_key

    @property
    def allowed_keys(self):
        """
        Returns the allowed keys in list format.
        """
        return self._allowed_keys

    async def get_neo_feed(self, options: Dict[str, Any]) -> NeoWsFeedResponse:
        """
        |coro|
        
        Same this as
        :py:class:`SyncNeoWs.get_neo_feed <nasawrapper.neows.SyncNeoWs.get_neo_feed>`,
        but with asynchronous syntax.

        **Example**

            .. code-block:: python3

                from nasawrapper import AsyncNeoWs
                from datetime import datetime
                import asyncio

                async def main():
                    neows = AsyncNeoWs("DEMO_KEY")
                    result = await neows.get_neo_feed({
                        "start_date": datetime(2010, 2, 3),
                        "end_date": datetime(2010, 2, 4)
                    })
                    print(result)

                loop = asyncio.get_event_loop()
                loop.run_until_complete(main())
        """
        url = f"https://api.nasa.gov/neo/rest/v1/feed?api_key={self._api_key}"
        options = Validator.validate(options, self._allowed_keys)

        # building url
        for key, value in options.items():
            url += f"&{key}={value}"

        # making request
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as request:
                if request.status == 429:
                    raise RateLimitError("You are being rate limited")
                elif request.status == 403:
                    raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

                request = await request.json()

        return request

    async def get_today_neo_feed(self) -> NeoWsFeedResponse:
        """
        |coro|
        
        Same this as
        :py:class:`SyncNeoWs.get_today_neo_feed <nasawrapper.neows.SyncNeoWs.get_today_neo_feed>`,
        but with asynchronous syntax.

        **Example**

            .. code-block:: python3

                from nasawrapper import AsyncNeoWs
                from datetime import datetime
                import asyncio

                async def main():
                    neows = AsyncNeoWs("DEMO_KEY")
                    result = await neows.get_today_neo_feed()
                    print(result)

                loop = asyncio.get_event_loop()
                loop.run_until_complete(main())
        """
        start_date = end_date = datetime.now()
        options = {
            "start_date": start_date,
            "end_date": end_date
        }
        return await self.get_neo_feed(options)

    async def get_neo_lookup(self, asteroid_id: int) -> Asteroid:
        """
        |coro|
        
        Lookup a specific asteroid based on its
        NASA JPL small body (SPK-ID) ID. Find more
        about this at
        `https://ssd.jpl.nasa.gov/tools/sbdb_query.html <https://ssd.jpl.nasa.gov/tools/sbdb_query.html>`_
        or simply use
        :py:class:`AsyncNeoWs.get_neo_browse <nasawrapper.neows.AsyncNeoWs.get_neo_browse>`
        to get the overall data-set of asteroids and their ids.

        **Example**

            .. code-block:: python3

                from nasawrapper import AsyncNeoWs
                from datetime import datetime
                import asyncio

                async def main():
                    neows = AsyncNeoWs("DEMO_KEY")
                    result = await neows.get_neo_lookup(3542519)
                    print(result)

                loop = asyncio.get_event_loop()
                loop.run_until_complete(main())
        """
        if not isinstance(asteroid_id, int):
            raise TypeError(f"'asteroid_id' must be 'int', got {asteroid_id.__class__.__name__}")

        url = f"https://api.nasa.gov/neo/rest/v1/neo/{asteroid_id}?api_key={self._api_key}"
        
        # making request
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as request:
                if request.status == 429:
                    raise RateLimitError("You are being rate limited")
                elif request.status == 403:
                    raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")
                elif request.status == 404:
                    raise NotFound(f"Asteroid of id '{asteroid_id}' could not be found")

                request = await request.json()

        return request

    async def get_neo_browse(self) -> NeoWsBrowseResponse:
        """
        |coro|
        
        Browse the overall asteroid data-set.

        **Example**

            .. code-block:: python3

                from nasawrapper import AsyncNeoWs
                from datetime import datetime
                import asyncio

                async def main():
                    neows = AsyncNeoWs("DEMO_KEY")
                    result = await neows.get_neo_browse()
                    print(result)

                loop = asyncio.get_event_loop()
                loop.run_until_complete(main())
        """
        url = f"https://api.nasa.gov/neo/rest/v1/neo/browse?api_key={self._api_key}"

        # making request
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as request:
                if request.status == 429:
                    raise RateLimitError("You are being rate limited")
                elif request.status == 403:
                    raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

                request = await request.json()

        return request

class NeoWsQueryBuilder:
    """
    Retrieve a list of Asteroids based on
    their closest approach date to Earth,
    just like
    :py:class:`SyncNeoWs.get_neo_feed <nasawrapper.neows.SyncNeoWs.get_neo_feed>`,
    except by the fact that you are the person who will build the query.

    **Example**

        .. code-block:: python3

            from nasawrapper import NeoWsQueryBuilder
            from datetime import datetime

            builder = NeoWsQueryBuilder("DEMO_KEY")
            result = builder.set_start_date(datetime(2020, 2, 3)).set_end_date(datetime(2020, 2, 4)).get_feed()
            print(result)
    """
    def __init__(self, api_key: str, options = {}) -> None:
        self._api_key = api_key
        self._options = options

    @property
    def api_key(self):
        """
        Returns the API key.
        """
        return self._api_key

    @property
    def options(self):
        """
        Returns the options in dict format.
        """
        return self._api_key

    def set_start_date(self, start_date: datetime):
        """
        Set 'start_date' field to the options.
        """
        if not isinstance(start_date, datetime):
            raise TypeError(f"'start_date' must be 'datetime.datetime', got {start_date.__class.__name__}")

        self._options["start_date"] = start_date   
        return NeoWsQueryBuilder(self._api_key, self._options)     

    def set_end_date(self, end_date: datetime):
        """
        Set 'end_date' field to the options.
        """
        if not isinstance(end_date, datetime):
            raise TypeError(f"'end_date' must be 'datetime.datetime', got {end_date.__class.__name__}")

        self._options["end_date"] = end_date 
        return NeoWsQueryBuilder(self._api_key, self._options)

    def get_feed(self):
        """
        Make the request with the provided
        information.
        """
        options = Validator.validate(self._options, ["start_date", "end_date"])
        
        url = f"https://api.nasa.gov/neo/rest/v1/feed?api_key={self._api_key}"

        # building url
        for key, value in options.items():
            url += f"&{key}={value}"

        # making request
        request = requests.get(url)
        if request.status_code == 429:
            raise RateLimitError("You are being rate limited")
        elif request.status_code == 403:
            raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

        return request.json()


import requests
import aiohttp
from typing import Dict, Union, Optional, Any, List, TypedDict
from datetime import datetime

from .errors import InvalidKey, InvalidDate, InvalidApiKey, RateLimitError

class ApodResponse(TypedDict):
    copyright: Optional[str]
    date: str
    explanation: str
    hd_url: Optional[str]
    media_type: str
    service_version: str
    title: str
    url: str

class Validator:
    """
    Will be called to validate the options provided
    by the developer
    """
    @classmethod
    def validate(
        cls,
        options: Dict[str, Any],
        allowed_keys: Dict[str, type],
        date_related_keys: List[str]
    ) -> Dict[str, Union[str, int, bool, datetime]]:
        """
        Function to validate the options acording to
        the API requisites
        """

        # creating checks for date related keys
        checks = [key in options for key in date_related_keys]

        # checking if the developer didn't provide any values
        if not any([key in options for key in allowed_keys]):
            raise ValueError("Missing properties like 'start_date' or 'end_date'.")

        # checking for 'end_date' and 'start_date'
        # 'start_date' can be used without 'end_date'
        # so ill not check this
        if checks[2] and not checks[1]:
            raise InvalidKey("'end_date' can not be used without 'start_date'")
        
        # checking if 'date' is being used with
        # 'start_date' or 'end_date'
        elif checks[0] and checks[2]:
            raise InvalidKey("'date' can not be used with 'end_date'")

        # interating over date keys to check their
        # types
        for key, value in allowed_keys.items():
            if options.get(key) and not isinstance(options[key], value):
                raise TypeError(f"'{key}' must be '{value.__name__}', got '{options[key].__class__.__name__}'")

        # validating dates
        for key in date_related_keys:
            if options.get(key):
                value = options[key]
                if value > datetime.now():
                    raise InvalidDate(f"'{key}' must be a valid date")

        # validating 'end_date' and 'start_date'
        if "end_date" in options.keys():
            if options["end_date"] < datetime(year=1995, month=6, day=16):
                raise InvalidDate("'end_date' must be after Jun 16, 1995.")

            elif options["end_date"] < options["start_date"]:
                raise InvalidDate(f"'end_date' can not be before 'start_date'")

            options["end_date"] = options["end_date"].strftime("%Y-%m-%d")
        
        if "start_date" in options.keys():
            options["start_date"] = options["start_date"].strftime("%Y-%m-%d")

        # checking 'count' and 'date' keys
        if "count" in options.keys() and any(checks):
            raise InvalidKey("'count' can not be used with 'end_date', 'start_date' or 'date'")

        if checks[0] in options.keys() and "start_date" in options.keys():
            raise InvalidKey("'start_date' can not be used with 'date'")

        return options
        

class SyncApod:
    """
    This class uses synchronous programming
    syntax to make requests to the
    ``APOD`` API and returns the API response.

    **Parameters**

        **api_key** (str) - The API key.
    """
    def __init__(self, api_key: str) -> None:
        self._api_key = api_key
        self._allowed_keys = {
            "date": datetime,
            "start_date": datetime,
            "end_date": datetime,
            "count": int,
            "thumbs": bool
        }
        self._date_related_keys = list(filter(lambda item: "date" in item, self._allowed_keys.keys()))
        self._base_url = f"https://api.nasa.gov/planetary/apod?api_key={self._api_key}"

    @property
    def api_key(self):
        """
        Returns the API key.
        """
        return self._api_key

    @property
    def allowed_keys(self):
        """
        Returns the allowed keys in a dict.
        """
        return self._allowed_keys

    @property
    def date_related_keys(self):
        """
        Returns the allowed keys in a dict.
        """
        return self._date_related_keys

    @property
    def base_url(self):
        """
        Returns the base URL.
        """
        return self._base_url

    def get_apod(self, options: Dict[str, Union[str, int, bool, datetime]]) -> Union[ApodResponse, List[ApodResponse]]:
        """
        Validate the provided options by checking their types
        and values. Here's a list of the allowed keys and an 
        example of how to call the method correctly:

        **Allowed Keys**

            ========== ============================= ==========================================================================================================
            Key        Type                          Function
            ========== ============================= ==========================================================================================================
            date       :py:class:`datetime.datetime` Search for a specific date. Can not be used with 'start_date'.
            start_date :py:class:`datetime.datetime` The start of a date range.
            end_date   :py:class:`datetime.datetime` The end of a date range. Can not be used without 'start_date' and can not be before Jun 16, 1995.
            count      :py:class:`int`               If specified, returns ``count`` randomly images. Can not be used with 'date', 'start_date' or 'end_date'.
            thumbs     :py:class:`bool`              If the APOD is a video, return the URL of the video. Default is ``False``.
            ========== ============================= ==========================================================================================================

        **Examples**

            **Getting asteroids from a range**

                .. code-block:: python3

                    from nasawrapper import SyncApod
                    from datetime import datetime, timedelta

                    apod = SyncApod("DEMO_KEY")
                    result = apod.get_apod({
                        "start_date": datetime.now() - timedelta(days=1),
                        "end_date": datetime.now(),
                        "thumbs": True
                    })

                    print(result)
            
            **Getting asteroids from specific date**
            
                .. code-block:: python3

                    from nasawrapper import SyncApod
                    from datetime import datetime, timedelta

                    apod = SyncApod("DEMO_KEY")
                    result = apod.get_apod({
                        "date": datetime(year=2010, month=3, day=2)
                    })

                    print(result)
        """
        options = Validator.validate(options, self._allowed_keys, self._date_related_keys)

        # building query
        for key, value in options.items():
            self._base_url += f"&{key}={value}"

        # making request
        request = requests.get(self._base_url)
        if request.status_code == 429:
            raise RateLimitError("You are being rate limited")
        elif request.status_code == 403:
            raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

        return request.json()

    def get_random(self) -> ApodResponse:
        """
        Returns a random picture of APOD API.
        You can manually do this by typing:

        .. code-block:: python3

            from nasawrapper import SyncApod

            apod = SyncApod("DEMO_KEY")
            result = apod.get_apod({
                "count": 1
            })
            print(result) # random picture

        But it's not recommended, since there's
        a specific method for this.
        """
        url = f"https://api.nasa.gov/planetary/apod?api_key={self._api_key}&count=1"

        # making request
        request = requests.get(url)
        if request.status_code == 429:
            raise RateLimitError("You are being rate limited")
        elif request.status_code == 403:
            raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

        return request.json()[0]

    def get_today_apod(self) -> ApodResponse:
        """
        Returns today's APOD. You can also
        clone this method manually by typing:

        .. code-block:: python3

            from nasawrapper import SyncApod
            from datetime import datetime

            apod = SyncApod("DEMO_KEY")
            result = apod.get_apod({
                "date": datetime.now()
            })
            print(result)

        But, for the same reasons as
        :py:class:`SyncApod.get_random <nasawrapper.apod.SyncApod.get_random>`,
        it's not recommended
        to do that.
        """
        now = datetime.now().strftime("%Y-%m-%d")
        url = f"https://api.nasa.gov/planetary/apod?api_key={self._api_key}&date={now}"
        
        # making request
        request = requests.get(url)
        if request.status_code == 429:
            raise RateLimitError("You are being rate limited")
        elif request.status_code == 403:
            raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

        return request.json()

        

        
class AsyncApod:
    """
    This class uses asynchronous programming
    syntax to make requests to the APOD API
    and returns the API response.

    **Parameters**
        
        **api_key** (str) - The API key.
    """
    def __init__(self, api_key: str) -> None:
        self._api_key = api_key
        self._allowed_keys = {
            "date": datetime,
            "start_date": datetime,
            "end_date": datetime,
            "count": int,
            "thumbs": bool
        }
        self._date_related_keys = list(filter(lambda item: "date" in item, self._allowed_keys))
        self._base_url = f"https://api.nasa.gov/planetary/apod?api_key={self._api_key}"

    @property
    def api_key(self):
        """
        Returns the API key.
        """
        return self._api_key

    @property
    def allowed_keys(self):
        """
        Returns the allowed keys in a dict.
        """
        return self._allowed_keys

    @property
    def date_related_keys(self):
        """
        Returns the allowed keys in a dict.
        """
        return self._date_related_keys

    @property
    def base_url(self):
        """
        Returns the base URL.
        """
        return self._base_url

    async def get_apod(self, options: Dict[str, Union[str, int, bool, datetime]]) -> Union[ApodResponse, List[ApodResponse]]:
        """
        |coro|

        Same thing as
        :py:class:`SyncApod.get_apod <nasawrapper.apod.SyncApod.get_apod>`,
        but with asynchronous syntax.

        **Example**

            .. code-block:: python3

                from nasawrapper import AsyncApod
                from datetime import datetime
                import asyncio

                async def main():
                    apod = AsyncApod("DEMO_KEY")
                    result = await apod.get_apod({
                        "date": datetime(2010, 2, 3)
                    })
                    print(result)

                loop = asyncio.get_event_loop()
                loop.run_until_complete(main())
        """
        options = Validator.validate(options, self._allowed_keys, self._date_related_keys)

        # building url
        for key, value in options.items():
            self._base_url += f"&{key}={value}"

        # making request
        async with aiohttp.ClientSession() as session:
            async with session.get(self._base_url) as response:
                if response.status == 429:
                    raise RateLimitError("You are being rate limited")
                elif response.status == 403:
                    raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")
                    
                response = await response.json()

        return response

    async def get_random(self) -> ApodResponse:
        """
        |coro|

        Same thing as
        :py:class:`SyncApod.get_random <nasawrapper.apod.SyncApod.get_random>`
        but with asynchronous syntax.

        **Example**

            .. code-block:: python3

                from nasawrapper import AsyncApod
                import asyncio

                async def main():
                    apod = AsyncApod("DEMO_KEY")
                    result = await apod.get_random()
                    print(result)

                loop = asyncio.get_event_loop()
                loop.run_until_complete(main())
        """
        url = f"https://api.nasa.gov/planetary/apod?api_key={self._api_key}&count=1"

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status == 429:
                    raise RateLimitError("You are being rate limited")
                elif response.status == 403:
                    raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")
                    
                response = await response.json()

        return response[0]

    async def get_today_apod(self) -> ApodResponse:
        """
        |coro|

        Same this as
        :py:class:`SyncApod.get_today_apod <nasawrapper.apod.SyncApod.get_today_apod>`
        but with asynchronous syntax.

        **Example**

            .. code-block:: python3

                from nasawrapper import AsyncApod
                import asyncio

                async def main():
                    apod = AsyncApod("DEMO_KEY")
                    result = await apod.get_today_apod()
                    print(result)

                loop = asyncio.get_event_loop()
                loop.run_until_complete(main())
        """
        now = datetime.now().strftime("%Y-%m-%d")
        url = f"https://api.nasa.gov/planetary/apod?api_key={self._api_key}&date={now}"

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status == 429:
                    raise RateLimitError("You are being rate limited")
                elif response.status == 403:
                    raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

                response = await response.json()

        return response

class ApodQueryBuilder:
    """
    If you want to build a query
    using methods, this wrapper
    provides exactly what you want.

    This class will build a query with
    the information provided by the methods.

    **Examples**

        **Getting 2 random pictures**

            .. code-block:: python3

                from nasawrapper import ApodQueryBuilder

                builder = ApodQueryBuilder("DEMO_KEY")
                result = builder.set_count(2).get_apod()
                print(result)

        **Getting picture from specific date**

            .. code-block:: python3

                from nasawrapper import ApodQueryBuilder
                from datetime import datetime

                builder = ApodQueryBuilder("DEMO_KEY")
                result = builder.set_date(datetime(2010, 2, 3))
                print(result)
    """
    def __init__(self, api_key: str, options = {}):
        self._api_key = api_key
        self._options = options

    @property
    def api_key(self):
        """
        Returns the API key.
        """
        return self._api_key

    @property
    def options(self):
        """
        Returns the options in dict format.
        """
        return self._options

    def set_date(self, date: datetime):
        """
        Add 'date' field to the options
        """
        if not isinstance(date, datetime):
            raise TypeError(f"'date' must be an 'datetime.datetime', got '{date.__class__.__name__}'")

        # checks whenever 'date' appears with
        # 'start_date' or 'count'
        if self._options.get("start_date"):
            raise InvalidKey("'date' can not be used with 'start_date'")    
        elif self._options.get("count"):
            raise InvalidKey("'date' can not be used with 'count'")

        self._options["date"] = date.strftime("%Y-%m-%d")

        return ApodQueryBuilder(self._api_key, self._options)

    def set_start_date(self, start_date: datetime):
        """
        Add 'start_date' field to the options
        """
        if not isinstance(start_date, datetime):
            raise TypeError(f"'start_date' must be an 'datetime.datetime', got '{start_date.__class__.__name__}'")

        if start_date < datetime(year=1995, month=6, day=16):
            raise InvalidDate("'end_date' must be after Jun 16, 1995.")

        return ApodQueryBuilder(self._api_key, self._options)

    def set_end_date(self, end_date: datetime):
        """
        Add 'end_date' field to the options
        """

        if not isinstance(end_date, datetime):
            raise TypeError(f"'end_date' must be an 'datetime.datetime', got '{end_date.__class__.__name__}'")

        return ApodQueryBuilder(self._api_key, self._options)

    def set_count(self, count: int):
        """
        Add 'count' field to the options
        """
        if not isinstance(count, int):
            raise TypeError(f"'count' must be 'int', got '{count.__class__.__name__}'")
        
        # checking whenever 'count' is being
        # used with 'date', 'start_date', 'end_date'
        checks = ["date" in self._options.keys(), "start_date" in self._options.keys(), "end_date" in self._options.keys()]
        if any(checks):
            raise InvalidKey("'count' can not be used with 'date', 'start_date' or 'end_date'")

        self._options["count"] = count
        return ApodQueryBuilder(self._api_key, self._options)

    def set_thumbs(self, thumbs: bool):
        """
        Add 'thumbs' field to the options.
        """
        if not isinstance(thumbs, bool):
            raise TypeError(f"'thumbs' must be 'bool', got '{thumbs.__class__.__name__}'")

        self._options["thumbs"] = thumbs
        return ApodQueryBuilder(self._api_key, self._options)

    def get_apod(self) -> Union[ApodResponse, List[ApodResponse]]:
        """
        Make the request with the provided
        information.
        """
        url = f"https://api.nasa.gov/planetary/apod?api_key={self._api_key}"

        options = Validator.validate(
            self._options,
            {
            "date": datetime,
            "start_date": datetime,
            "end_date": datetime,
            "count": int,
            "thumbs": bool
            },
            ["date", "start_date", "end_date"]
        )

        # building query
        for key, value in options.items():
            url += f"&{key}={value}"

        # making request
        request = requests.get(url)
        if request.status_code == 429:
            raise RateLimitError("You are being rate limited")
        elif request.status_code == 403:
            raise InvalidApiKey(f"'{self._api_key}' is not a valid API key")

        return request.json()


class NotFound(Exception):
    """
    Exception that's raised when an item is not found in
    any NASA API
    """
    pass

class InvalidKey(Exception):
    """
    Exception that's raised when the developer provides
    an invalid key
    """
    pass

class InvalidDate(Exception):
    """
    Exception that's raised when the developer provides
    an invalid date
    """
    pass

class InvalidApiKey(Exception):
    """
    Exception that's raised when the developer
    provides an invalid API key
    """
    pass

class RateLimitError(Exception):
    """
    Exception that's raised when the develper
    make too many requests to the API
    """


from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as rm:
    readme = rm.read()

requirements = []
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="nasawrapper",
    author="End313234",
    url="https://github.com/End313234/nasawrapper-python",
    project_urls={
        "Bug Hunter": "https://github.com/End313234/nasawrapper/issues",
        "Documentation": "https://github.com/End313234/nasawrapper#documentation"
    },
    version="0.3.1",
    license="MIT",
    description="A simple wrapper to fetch NASA Open APIs using Python",
    long_description=readme,
    long_description_content_type="text/markdown",
    python_requires=">=3.6.0",
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Topic :: Internet",
        "Development Status :: 3 - Alpha"
    ],
    packages=find_packages(),
    install_requires=requirements
)


import requests

def get_remaining_rate_limit(api_key: str) -> int:
    """
    Returns your remaining rate limit by
    making a request to
    :ref:`Apod <extensions/apod:Apod>` and
    getting the header ``X-RateLimit-Remaining``,
    that's returned on every API response.

    For example, if you are using an
    API key different from ``DEMO_KEY``,
    you have a default hourly rate
    limit of 1.000 requests, acording
    to the `Portal <https://api.nasa.gov/>`_.
    So, if you make 2 requests, your
    remaining rate limit will be equal to
    998.

    **Example**
        
        .. code-block:: python3

            from nasawrapper.utils import get_remaining_rate_limit

            remaining = get_remaining_rate_limit("DEMO_KEY")
            print(reamining)
    """
    headers = requests.get(f"https://api.nasa.gov/planetary/apod?api_key={api_key}").headers
    return int(headers["X-RateLimit-Remaining"])